"""Classes used by the Inputs class"""
import os
import numpy as np
import pickle
import pandas as pd
from astropy.time import Time
import astropy.units as u
import sqlalchemy as sqla
import sqlalchemy.dialects.postgresql as pg
from nexoclom.solarsystem import SSObject
from nexoclom.utilities import InputError, NexoclomConfig


dtor = np.pi/180.
# Tolerances for floating point values
dtaa = 2.*dtor


config = NexoclomConfig()


class Geometry:
    def __init__(self, gparam):
        """Define a Geometry object.
        
        See :doc:`inputfiles#Geometry` for more information.
        """
        # Planet
        planet = gparam.get('planet', None)
        if planet is None:
            raise InputError('Geometry.__init__',
                             'Planet not defined in inputfile.')
        else:
            self.planet = SSObject(planet.title())

        # All possible objects
        objlist = [self.planet.object]
        if self.planet.moons is not None:
            objlist.extend([m.object for m in self.planet.moons])
        else:
            pass

        # Choose the starting point
        self.startpoint = gparam.get('startpoint', self.planet.object).title()
        if self.startpoint not in objlist:
            print(f'{self.startpoint} is not a valid starting point.')
            olist = '\n\t'.join(objlist)
            print(f'Valid choices are:\n\t{olist}')
            raise ValueError
        else:
            pass

        # Choose which objects to include
        # This is given as a list of names
        # Default = geometry.planet and geometry.startpoint
        if 'objects' in gparam:
            inc = set(i.strip().title()
                      for i in gparam['objects'].split(','))
        else:
            inc = {self.planet.object, self.startpoint}

        for i in inc:
            if i not in objlist:
                raise InputError('Geometry.__init__',
                                 f'Invalid object {i} in geometry.include')
            
        # Only remember objects that will be included
        self.objects = set(SSObject(o) for o in inc)
        if len(self.objects) == 0:
            # Probably not possible to get here
            self.objects = None
        else:
            pass

        # Different objects are created for geometry_with_starttime and
        # geometry_without_starttime
        if 'starttime' in gparam:
            self.type = 'geometry with starttime'
            try:
                self.time = Time(gparam['starttime'].upper())
            except:
                raise InputError('Geometry.__init__',
                    f'Invalid starttime format: {gparam["starttime"]}')
        else:
            self.type = 'geometry without starttime'
            if len(self.planet) == 1:
                self.phi = None
            elif 'phi' in gparam:
                phi_ = gparam['phi'].split(',')
                phi = tuple(float(p)*u.rad for p in phi_)
                nmoons = len(self.objects - {self.planet})
                if len(phi) == nmoons:
                    self.phi = phi
                else:
                    raise InputError('Geometry.__init__',
                        'The wrong number of orbital positions was given.')
            else:
                raise InputError('Geometry.__init__',
                    'geometry.phi was not specified.')

            # Subsolar longitude and latitude
            if 'subsolarpoint' in gparam:
                subs = gparam['subsolarpoint'].split(',')
                try:
                    self.subsolarpoint = (float(subs[0])*u.rad,
                                          float(subs[1])*u.rad)
                except:
                    raise InputError('Geometry.__init__',
                        'The format for geometry.subsolarpoint is wrong.')
            else:
                self.subsolarpoint = (0*u.rad, 0*u.rad)

            # True Anomaly Angle
            self.taa = float(gparam.get('taa', 0.))*u.rad
            
    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
        
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'geometry.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        # check to see if it is already there
        ids = self.search()
        
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            if self.type == 'geometry with starttime':
                if self.objects is None:
                    objs = None
                else:
                    objs = [o.object for o in self.objects]

                table = sqla.Table("geometry_without_time",
                                   metadata_obj,
                                   autoload_with=engine)
                insert_stmt = pg.insert(table).values(
                    planet=self.planet.object,
                    startpoint=self.startpoint,
                    objects=objs,
                    starttime = self.time)
            elif self.type == 'geometry without starttime':
                if self.objects is None:
                    objs = None
                else:
                    objs = [o.object for o in self.objects]
                    
                subspt = [s.value for s in self.subsolarpoint]
                
                if self.phi is None:
                    phi = None
                else:
                    phi = [p.value for p in self.phi]
                    
                table = sqla.Table("geometry_without_time",
                                   metadata_obj,
                                   autoload_with=engine)
                
                insert_stmt = pg.insert(table).values(
                    planet=self.planet.object,
                    startpoint=self.startpoint,
                    objects=objs,
                    phi=phi,
                    subsolarpt=subspt,
                    taa=self.taa.value)
            else:
                raise InputError('Geometry.insert()',
                                 f'geometry.type = {self.type} not allowed.')
            
            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()
            
            ids = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass
        
        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()

        if self.objects is None:
            objects = None
        else:
            objects = [o.object for o in self.objects]
            
        if self.type == 'geometry with starttime':
            table = sqla.Table("geometry_without_time",
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum,
                                table.columns.taa).where(
                table.columns.planet == self.planet.object,
                table.columns.startpoint == self.startpoint,
                table.columns.objects  ==  objects,
                table.columns.starttime  ==  self.time)
        elif self.type == 'geometry without starttime':
            if self.phi is None:
                phi = None
            else:
                phi = [p.value for p in self.phi]
            
            subspoint = [s.value for s in self.subsolarpoint]

            table = sqla.Table("geometry_without_time",
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum,
                                table.columns.taa).where(
                table.columns.planet == self.planet.object,
                table.columns.startpoint == self.startpoint,
                table.columns.objects == objects,
                table.columns.phi == phi,
                table.columns.subsolarpt == subspoint,
                table.columns.taa >= self.taa.value - dtaa/2.,
                table.columns.taa < self.taa.value + dtaa/2.)
        else:
            raise InputError('geometry.search()',
                             f'geometry.type = {self.type} not allowed.')
                    
        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))
            
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            diff = np.abs(results.taa - self.taa.value)
            q = np.where(diff == diff.min())[0]
            if len(q) == 1:
                return int(results.loc[q, 'idnum'][0])
            else:
                raise RuntimeError('geometry.search()',
                                   'Duplicates in geometry table')


class SurfaceInteraction:
    def __init__(self, sparam):
        """Define a SurfaceInteraction object.

        See :doc:`inputfiles#SurfaceInteraction` for more information.
        """
        sticktype = (sparam['sticktype'].lower()
                     if 'sticktype' in sparam
                     else None)
        if sticktype == 'temperature dependent':
            self.sticktype = sticktype
            
            if 'accomfactor' in sparam:
                self.accomfactor = float(sparam['accomfactor'])
            else:
                raise InputError('SurfaceInteraction.__init__',
                                 'surfaceinteraction.accomfactor not given.')
            
            if 'a' in sparam:
                A = tuple([float(a) for a in sparam['a'].split(',')])
                if len(A) == 3:
                    self.A = A
                else:
                    raise InputError('SurfaceInteraction.__init__',
                                     'surfaceinteraction.A must have 3 values')
            else:
                self.A = (1.57014, -0.006262, 0.1614157)
        elif sticktype == 'surface map':
            self.sticktype = sticktype
            self.stick_mapfile = sparam.get('stick_mapfile', 'default')

            if os.path.exists(self.stick_mapfile):
                if self.stick_mapfile.endswith('.pkl'):
                    with open(self.stick_mapfile, 'rb') as f:
                        sourcemap = pickle.load(f)
                    self.coordinate_system = sourcemap.get('coordinate_system',
                                                           'solar-fixed')
                else:
                    self.coordinate_system = 'solar-fixed'
                    # assert 0, 'Not set up yet'
            else:
                print('Warning: mapfile does not exist')
                self.coordinate_system = 'solar-fixed'

            if (self.coordinate_system == 'planet-fixed'):
                self.subsolarlon = sparam.get('subsolarlon', 0.) * u.rad
            else:
                self.subsolarlon = None

            if 'accomfactor' in sparam:
                self.accomfactor = float(sparam['accomfactor'])
            else:
                raise InputError('SurfaceInteraction.__init__',
                                 'surfaceinteraction.accomfactor not given.')
        elif 'stickcoef' in sparam:
            # Constant sticking
            self.sticktype = 'constant'
            self.stickcoef = float(sparam['stickcoef'])
            if self.stickcoef < 0:
                self.stickcoef = 0
            elif self.stickcoef > 1:
                self.stickcoef = 1
            else:
                pass
            
            if 'accomfactor' in sparam:
                self.accomfactor = float(sparam['accomfactor'])
            else:
                if self.stickcoef == 1:
                    self.accomfactor = None
                else:
                    raise InputError('SurfaceInteraction.__init__',
                                 'surfaceinteraction.accomfactor not given.')
        else:
            self.sticktype = 'constant'
            self.stickcoef = 1.
            self.accomfactor = None

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'surfaceinteraction.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            if self.sticktype == 'constant':
                table = sqla.Table('surface_int_constant',
                                   metadata_obj,
                                   autoload_with=engine)
                
                insert_stmt = pg.insert(table).values(
                    stickcoef=self.stickcoef,
                    accomfactor=self.accomfactor)
            elif self.sticktype == 'surface map':
                table = sqla.Table('surface_int_map',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    mapfile=self.stick_mapfile,
                    accomfactor = self.accomfactor)
            elif self.sticktype == 'temperature dependent':
                table = sqla.Table('surface_int_tempdependent',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    accomfactor = self.accomfactor,
                    a = self.A)
            else:
                raise InputError('SurfaceInteraction.search()',
                    f'surfaceinteraction.sticktype = {self.sticktype} not allowed.')

            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()

            ids = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass

        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()
        
        if self.sticktype == 'constant':
            table = sqla.Table('surface_int_constant',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.accomfactor == self.accomfactor,
                table.columns.stickcoef == self.stickcoef)
        elif self.sticktype == 'temperature dependent':
            table = sqla.Table('surface_int_tempdependent',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.accomfactor == self.accomfactor,
                table.columns.a == self.A)
        elif self.sticktype == 'surface map':
            table = sqla.Table('surface_int_map',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.accomfactor == self.accomfactor,
                table.columns.mapfile == self.stick_mapfile)
        else:
            raise InputError('SurfaceInteraction.search()',
             f'surfaceinteraction.sticktype = {self.sticktype} not allowed.')
        
        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))
            
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            raise RuntimeError('SurfaceInteraction.search()',
                               'Duplicates in surface interaction table')


class Forces:
    def __init__(self, fparam):
        """Define a Forces object.

        See :doc:`inputfiles#Forces` for more information.
        """
    
        self.gravity = (bool(eval(fparam['gravity'].title()))
                        if 'gravity' in fparam
                        else True)
        self.radpres = (bool(eval(fparam['radpres'].title()))
                        if 'radpres' in fparam
                        else True)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'forces.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            table = sqla.Table('forces', metadata_obj, autoload_with=engine)
            insert_stmt = pg.insert(table).values(
                gravity=self.gravity,
                radpres=self.radpres)
            
            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()
                
            ids  = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass

        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()
        table = sqla.Table('forces', metadata_obj, autoload_with=engine)
        
        query = sqla.select(table.columns.idnum).where(
            table.columns.gravity == self.gravity,
            table.columns.radpres == self.radpres)
        
        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))
    
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            raise RuntimeError('Forces.search()', 'Duplicates in forces table')


class SpatialDist:
    def __init__(self, sparam):
        """Define a SpatialDist object.

        See :doc:`inputfiles#SpatialDist` for more information.
        """
        if 'type' in sparam:
            self.type = sparam['type']
        else:
            raise InputError('SpatialDist.__init__',
                             'SpatialDist.type not given')
        
        if self.type == 'uniform':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            if 'longitude' in sparam:
                lon0, lon1 = (float(l.strip())
                              for l in sparam['longitude'].split(','))
                lon0 = max(lon0, 0.)
                lon0 = min(lon0, 2*np.pi)
                lon1 = max(lon1, 0.)
                lon1 = min(lon1, 2*np.pi)
                self.longitude = (lon0*u.rad, lon1*u.rad)
            else:
                self.longitude = (0.*u.rad, 2*np.pi*u.rad)
                
            if 'latitude' in sparam:
                lat0, lat1 = (float(l.strip())
                              for l in sparam['latitude'].split(','))
                lat0 = max(lat0, -np.pi/2)
                lat0 = min(lat0, np.pi/2)
                lat1 = max(lat1, -np.pi/2)
                lat1 = min(lat1, np.pi/2)
                if lat0 > lat1:
                    raise InputError('SpatialDist.__init__',
                         'SpatialDist.latitude[0] > SpatialDist.latitude[1]')
                self.latitude = (lat0*u.rad, lat1*u.rad)
            else:
                self.latitude = (-np.pi/2*u.rad, np.pi/2*u.rad)
        elif self.type == 'surface map':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            
            self.mapfile = sparam.get('mapfile', 'default')
            
            if os.path.exists(self.mapfile):
                if self.mapfile.endswith('.pkl'):
                    with open(self.mapfile, 'rb') as f:
                        sourcemap = pickle.load(f)
                    coord_systems = ['planet-fixed', 'solar-fixed']
                    assert sourcemap['coordinate_system'] in coord_systems, (
                        'source map must be in planet-fixed or solar-fixed coordinates.')
                    self.coordinate_system = sourcemap['coordinate_system']
                else:
                    self.coordinate_system = 'solar-fixed'
                    # assert 0, 'Not set up yet'
            else:
                print('Warning: mapfile does not exist')
                self.coordinate_system = 'planet-fixed'

            if (self.coordinate_system == 'planet-fixed'):
                self.subsolarlon = sparam.get('subsolarlon', 0.)*u.rad
            else:
                self.subsolarlon = None
        elif self.type == 'surface spot':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            if 'longitude' in sparam:
                self.longitude = float(sparam['longitude'])*u.rad
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpatialDist.longitude not given.')
            
            if 'latitude' in sparam:
                self.latitude = float(sparam['latitude'])*u.rad
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpatialDist.latitude not given.')

            if 'sigma' in sparam:
                self.sigma = float(sparam['sigma'])*u.rad
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpatialDist.sigma not given.')
        elif self.type == 'fitted output':
            self.unfit_outid = -1
            self.query = None
        else:
            raise InputError('SpatialDist.__init__',
                             f'SpatialDist.type = {self.type} not defined.')

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'SpatialDist.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            if self.type == 'uniform':
                long = [l.value for l in self.longitude]
                lat = [l.value for l in self.latitude]
                
                table = sqla.Table("spatdist_uniform",
                                   metadata_obj,
                                   autoload_with=engine)
                insert_stmt = pg.insert(table).values(
                    exobase = self.exobase,
                    longitude = long,
                    latitude = lat)
            elif self.type == 'surface map':
                sslon = (None
                         if self.subsolarlon is None
                         else self.subsolarlon.value)
                table = sqla.Table("spatdist_surfmap",
                                   metadata_obj,
                                   autoload_with=engine)
                insert_stmt = pg.insert(table).values(
                    exobase=self.exobase,
                    mapfile=self.mapfile,
                    subsolarlon=sslon,
                    coordinate_system=self.coordinate_system)
            elif self.type == 'surface spot':
                table = sqla.Table("spatdist_spot",
                                   metadata_obj,
                                   autoload_with=engine)
                insert_stmt = pg.insert(table).values(
                    exobase=self.exobase,
                    longitude=self.longitude.value,
                    latitude=self.latitude.value,
                    sigma=self.sigma.value)
            elif self.type == 'fitted output':
                table = sqla.Table("spatdist_fittedoutput",
                                   metadata_obj,
                                   autoload_with=engine)
                insert_stmt = pg.insert(table).values(
                    unfit_outid=self.unfit_outid,
                    query=self.query)
            else:
                raise InputError('SpatialDist.search()',
                                 f'SpatialDist.type = {self.type} not allowed.')
            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()

            ids  = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass
        
        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()
        
        if self.type == 'uniform':
            table = sqla.Table('spatdist_uniform',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.exobase == self.exobase,
                table.columns.longitude ==  [self.longitude[0].value,
                                             self.longitude[1].value],
                table.columns.latitude == [self.latitude[0].value,
                                           self.latitude[1].value])
        elif self.type == 'surface map':
            sslon = (None
                     if self.subsolarlon is None
                     else self.subsolarlon.value)
            table = sqla.Table('spatdist_surfmap',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.exobase == self.exobase,
                table.columns.mapfile == self.mapfile,
                table.columns.subsolarlon == sslon,
                table.columns.coordinate_system == self.coordinate_system)
        elif self.type == 'surface spot':
            table = sqla.Table('spatdist_spot',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.exobase == self.exobase,
                table.columns.longitude == self.longitude.value,
                table.columns.latitude == self.latitude.value,
                table.columns.sigma == self.sigma.value)
        elif self.type == 'fitted output':
            table = sqla.Table('spatdist_fittedoutput',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.unfit_outid == self.unfit_outid,
                table.columns.query == self.query)
        else:
            raise InputError('SpatialDist.__init__',
                             f'SpatialDist.type = {self.type} not defined.')

        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))
    
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            raise RuntimeError('SpatialDist.search()',
                               'Duplicates in forces table')


class SpeedDist:
    """Define a SpeedDist object.

    See :doc:`inputfiles#SpeedDist` for more information.
    """
    def __init__(self, sparam):
        self.type = sparam['type']

        if self.type == 'gaussian':
            if 'vprob' in sparam:
                self.vprob = float(sparam['vprob'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.vprob not given.')
            if 'sigma' in sparam:
                self.sigma = float(sparam['sigma'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.sigma not given.')
        elif self.type == 'sputtering':
            if 'alpha' in sparam:
                self.alpha = float(sparam['alpha'])
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.alpha not given.')
            if 'beta' in sparam:
                self.beta = float(sparam['beta'])
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.beta not given.')
            if 'u' in sparam:
                self.U = float(sparam['u'])*u.eV
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.U not given.')
        elif self.type == 'maxwellian':
            if 'temperature' in sparam:
                self.temperature = float(sparam['temperature'])*u.K
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.temperature not given.')
        elif self.type == 'flat':
            if 'vprob' in sparam:
                self.vprob = float(sparam['vprob'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.vprob not given.')
            
            if 'delv' in sparam:
                self.delv = float(sparam['delv'])*u.km/u.s
            else:
                raise InputError('SpatialDist.__init__',
                                 'SpeedDist.delv not given.')
        elif self.type == 'user defined':
            self.vdistfile = sparam.get('vdistfile', 'default')
        elif self.type == 'fitted output':
            self.unfit_outid = -1
            self.query = None
        else:
            assert 0, f'SpeedDist.type = {self.type} not available'

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'SpeedDist.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            if self.type == 'gaussian':
                table = sqla.Table('speeddist_gaussian',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    vprob=self.vprob.value,
                    sigma=self.sigma.value)
            elif self.type == 'sputtering':
                table = sqla.Table('speeddist_sputtering',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    alpha=self.alpha,
                    beta=self.beta,
                    U=self.U.value)
            elif self.type == 'maxwellian':
                table = sqla.Table('speeddist_maxwellian',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    temperature=self.temperature.value)
            elif self.type == 'flat':
                table = sqla.Table('speeddist_flat',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    vprob=self.vprob.value,
                    delv=self.delv.value)
            elif self.type == 'user defined':
                table = sqla.Table('speeddist_user',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    vdistfile=self.vdistfile)
            elif self.type == 'fitted output':
                table = sqla.Table('speeddist_fittedoutput',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    unfit_outid=self.unfit_outid,
                    query=self.query)
            else:
                raise InputError('SpeedDist.search()',
                                 f'speeddist.type = {self.type} not allowed.')

            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()

            ids = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass

        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()
        
        if self.type == 'gaussian':
            table = sqla.Table('vprob_gaussian',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.vprob == self.vprob.value,
                table.columns.sigma == self.sigma.value)
        elif self.type == 'sputtering':
            table = sqla.Table('speeddist_sputtering',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.alpha == self.alpha,
                table.columns.beta == self.beta,
                table.columns.U == self.U.value)
            
        elif self.type == 'maxwellian':
            table = sqla.Table('speeddist_maxwellian',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.temperature == self.temperature.value)
        elif self.type == 'flat':
            table = sqla.Table('speeddist_flat',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.vprob == self.vprob.value,
                table.columns.delv == self.delv.value)
        elif self.type == 'user defined':
            table = sqla.Table('speeddist_user',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.vdistfile == self.vdistfile)
        elif self.type == 'fitted output':
            table = sqla.Table('speeddist_fittedoutput',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.unfit_outid == self.unfit_outid,
                table.columns.query == self.query)
        else:
            raise InputError('SpeedDist.__init__',
                             f'SpeedDist.type = {self.type} not defined.')

        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))
            
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            raise RuntimeError('SpeedDist.search()',
                               'Duplicates in speed distribution table')
    

class AngularDist:
    def __init__(self, aparam):
        """Define a AngularDist object.

        See :doc:`inputfiles#AngularDist` for more information.
        """
        if 'type' in aparam:
            self.type = aparam['type']
            if self.type == 'radial':
                pass
            elif self.type == 'isotropic':
                if 'azimuth' in aparam:
                    az0, az1 = (float(l.strip())
                                for l in aparam['azimuth'].split(','))
                    az0 = max(az0, 0.)
                    az0 = min(az0, 2*np.pi)
                    az1 = max(az1, 0.)
                    az1 = min(az1, 2*np.pi)
                    self.aziumth = (az0*u.rad, az1*u.rad)
                else:
                    self.azimuth = (0*u.rad, 2*np.pi*u.rad)

                if 'altitude' in aparam:
                    alt0, alt1 = (float(l.strip())*u.rad
                                  for l in aparam['altitude'].split(','))
                    alt0 = max(alt0, 0)
                    alt0 = min(alt0, np.pi/2)
                    alt1 = max(alt1, 0)
                    alt1 = min(alt1, np.pi/2)
                    if alt0 > alt1:
                        raise InputError('AngularDist.__init__',
                         'AngularDist.altitude[0] > AngularDist.altitude[1]')
                    self.altitude = (alt0*u.rad, alt1*u.rad)
                else:
                    self.altitude = (0*u.rad, np.pi/2*u.rad)
            else:
                raise InputError('AngularDist.__init__',
                             f'AngularDist.type = {self.type} not defined.')
        else:
            self.type = 'isotropic'
            self.azimuth = (0*u.rad, 2*np.pi*u.rad)
            self.altitude = (0*u.rad, np.pi/2*u.rad)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'AngularDist.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            
            if self.type == 'radial':
                assert False, 'Should not be able to get here.'
            elif self.type == 'isotropic':
                alt = [a.value for a in self.altitude]
                az = [a.value for a in self.azimuth]
                table = sqla.Table('angdist_isotropic',
                                   metadata_obj,
                                   autoload_with=engine)
    
                insert_stmt = pg.insert(table).values(
                    altitude=alt,
                    azimuth=az)
            else:
                raise InputError('AngularDist.search()',
                                 f'angulardist.type = {self.type} not allowed.')

            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()

            ids = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass

        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()
        
        if self.type == 'radial':
            return 0
        elif self.type == 'isotropic':
            alt = [a.value for a in self.altitude]
            az = [a.value for a in self.azimuth]
            table = sqla.Table('angdist_isotropic',
                               metadata_obj,
                               autoload_with=engine)
            query = sqla.select(table.columns.idnum).where(
                table.columns.altitude == alt,
                table.columns.azimuth == az)
        else:
            raise InputError('AngularDist.__init__',
                             f'AngularDist.type = {self.type} not defined.')
    
        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))
            
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            raise RuntimeError('AngularDist.search()',
                               'Duplicates in angular distribution table')


class Options:
    def __init__(self, oparam):
        """Define a Options object.

        See :doc:`inputfiles#Options` for more information.
        """
        if 'endtime' in oparam:
            self.endtime = float(oparam['endtime'])*u.s
        else:
            raise InputError('Options.__init__',
                             'options.endtime not specified.')

        if 'species' in oparam:
            self.species = oparam['species'].capitalize()
        elif 'atom' in oparam:
            self.species = oparam['atom'].capitalize()
        else:
            raise InputError('Options.__init__',
                             'options.species not specified.')

        self.lifetime = float(oparam.get('lifetime', 0))*u.s
        
        if 'outeredge' in oparam:
            self.outeredge = float(oparam['outeredge'])
        elif 'outer_edge' in oparam:
            self.outeredge = float(oparam['outer_edge'])
        else:
            self.outeredge = 1e30
            
        if 'step_size' in oparam:
            self.step_size = float(oparam['step_size'])
        elif 'stepsize' in oparam:
            self.step_size = float(oparam['step_size'])
        else:
            self.step_size = 0.

        if self.step_size == 0:
            self.resolution = oparam.get('resolution', 1e-4)
        else:
            self.resolution = None
            
        if 'fitted' in oparam:
            self.fitted = (True if oparam['fitted'].casefold() == 'True'.casefold()
                           else False)
        else:
            self.fitted = False

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            pass
    
        keys_self, keys_other = set(self.__dict__.keys()), set(other.__dict__.keys())
        if keys_self != keys_other:
            return False
        else:
            return all([self.__dict__[key] == other.__dict__[key] for key in keys_self])

    def __str__(self):
        result = ''
        for key,value in self.__dict__.items():
            result += f'options.{key} = {value}\n'
        return result.strip()
    
    def insert(self):
        ids = self.search()
        if ids is None:
            engine = config.create_engine()
            metadata_obj = sqla.MetaData()
            table = sqla.Table('options',
                               metadata_obj,
                               autoload_with=engine)

            insert_stmt = pg.insert(table).values(
                endtime=self.endtime.value,
                species=self.species,
                lifetime=self.lifetime.value,
                outer_edge=self.outeredge,
                step_size=self.step_size,
                resolution=self.resolution,
                fitted=self.fitted)
                
            with engine.connect() as con:
                result = con.execute(insert_stmt)
                con.commit()

            ids = result.inserted_primary_key
            assert len(ids) == 1
            ids = ids[0]
        else:
            pass

        return int(ids)

    def search(self):
        engine = config.create_engine()
        metadata_obj = sqla.MetaData()
        
        table = sqla.Table('options',
                           metadata_obj,
                           autoload_with=engine)
        query = sqla.select(table.columns.idnum).where(
            table.columns.endtime == self.endtime.value,
            table.columns.species == self.species,
            table.columns.lifetime == self.lifetime.value,
            table.columns.outer_edge == self.outeredge,
            table.columns.step_size == self.step_size,
            table.columns.resolution == self.resolution,
            table.columns.fitted == self.fitted)
        
        with engine.connect() as con:
            results = pd.DataFrame(con.execute(query))

        if len(results) == 0:
            return None
        elif len(results) == 1:
            return int(results.loc[0, 'idnum'])
        else:
            raise RuntimeError('Options.search()', 'Duplicates in options table')



from nexoclom.utilities.NexoclomConfig import NexoclomConfig
from nexoclom.utilities.configure import configure
from nexoclom.utilities.exceptions import *


__name__ = 'utilities'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'



import numpy as np
import pytest
import astropy.units as u
from nexoclom.modelcode.source_distribution import xyz_from_lonlat


@pytest.mark.modelcode
def test_xyz_from_latlon():
    # Check points around the equator for planet
    lon = np.arange(0, 2*np.pi, np.pi/4)*u.rad
    lat = np.zeros_like(lon)
    exobase = 1.
    isplan = True
    xyz = xyz_from_lonlat(lon, lat, isplan, exobase)

    x = [0, np.sqrt(2)/2, 1., np.sqrt(2)/2, 0., -np.sqrt(2)/2, -1, -np.sqrt(2)/2]*u.rad
    y = [-1, -np.sqrt(2)/2, 0, np.sqrt(2)/2, 1, np.sqrt(2)/2, 0, -np.sqrt(2)/2]*u.rad
    z = np.zeros_like(x)*u.rad
    expected = np.array([x, y, z])
    
    assert xyz == pytest.approx(expected)
    
    # Check points along latitude lines
    lon = np.linspace(0, np.pi, 5)
    lat = np.linspace(-np.pi/2, np.pi/2, 5)
    exobase = 2.
    isplan = True
    xyz = xyz_from_lonlat(lon, lat, isplan, exobase)

    z = np.array([-1, -np.sqrt(2)/2, 0, np.sqrt(2)/2, 1])
    z_ = np.array([0, np.sqrt(2)/2, 1, np.sqrt(2)/2, 0])
    x = np.array([0, np.sqrt(2)/2, 1., np.sqrt(2)/2, 0]) * z_
    y = np.array([-1, -np.sqrt(2)/2, 0., np.sqrt(2)/2, 1]) * z_
    expected = np.array([x, y, z]) * exobase

    assert xyz == pytest.approx(expected)
    
    # Check points around the equator for moon
    lon = np.arange(0, 2*np.pi, np.pi/4)*u.rad
    lat = np.zeros_like(lon)
    exobase = 1.
    isplan = False
    xyz = xyz_from_lonlat(lon, lat, isplan, exobase)

    x = [0, -np.sqrt(2)/2, -1., -np.sqrt(2)/2, 0., np.sqrt(2)/2, 1, np.sqrt(2)/2]*u.rad
    y = [-1, -np.sqrt(2)/2, 0, np.sqrt(2)/2, 1, np.sqrt(2)/2, 0, -np.sqrt(2)/2]*u.rad
    z = np.zeros_like(x)*u.rad
    expected = np.array([x, y, z])
    
    assert xyz == pytest.approx(expected)
    
    # Check points along latitude lines
    lon = np.linspace(0, np.pi, 5)
    lat = np.linspace(-np.pi/2, np.pi/2, 5)
    exobase = 2.
    isplan = False
    xyz = xyz_from_lonlat(lon, lat, isplan, exobase)

    z = np.array([-1, -np.sqrt(2)/2, 0, np.sqrt(2)/2, 1])
    z_ = np.array([0, np.sqrt(2)/2, 1, np.sqrt(2)/2, 0])
    x = np.array([0, -np.sqrt(2)/2, -1., -np.sqrt(2)/2, 0]) * z_
    y = np.array([-1, -np.sqrt(2)/2, 0., np.sqrt(2)/2, 1]) * z_
    expected = np.array([x, y, z]) * exobase

    assert xyz == pytest.approx(expected)
    
    
if __name__ == '__main__':
    test_xyz_from_latlon()



"""Compute random deviates from arbitrary 1D and 2D distributions."""
import numpy as np
import numpy.random as random
from scipy import interpolate
import astropy.units as u


def random_deviates_1d(x, f_x, num):
    """Compute random deviates from arbitrary 1D distribution.

    f_x does not need to integrate to 1. The function nomralizes the
    distribution. Uses Transformation method (Numerical Recepies, 7.3.2)

    **Parameters**
    
    x
        The x values of the distribution

    f_x
        The relative probability of the value being in x and x+dx

    num
        The number of random deviates to compute

    **Returns**

    numpy array of length num chosen from the distribution f_x.
    """
    x_ = np.linspace(x.min(), x.max(), f_x.shape[0])
    cumsum = f_x.cumsum()
    cumsum -= cumsum.min()
    cumsum /= cumsum.max()
    return np.interp(random.rand(num), cumsum, x_)


def random_deviates_2d(fdist, x0, y0, num):
    """Compute random deviates from arbitrary 2D distribution.

    Uses acceptance/rejection method.
    **Parameters**
        fdist
            2d array of relative probability
        x0
            xaxis
        y0
            yaxis
        num
            number of points to choose
            
    **Outputs**
        x, y
            vectors of length num
    """
    mx = (x0.max()-x0.min(), x0.min())
    my = (y0.max()-y0.min(), y0.min())
    fmax = fdist.max()
    
    x0_ = np.linspace(x0.min(), x0.max(), fdist.shape[0])
    y0_ = np.linspace(y0.min(), y0.max(), fdist.shape[1])

    xpts, ypts = [], []
    while len(xpts) < num:
        ux = random.rand(num)*mx[0] + mx[1]
        uy = random.rand(num)*my[0] + my[1]
        uf = random.rand(num)*fmax

        val = interpolate.interpn((x0_, y0_), fdist, (ux, uy))
        mm = uf < val
        xpts.extend(list(ux[mm]))
        ypts.extend(list(uy[mm]))

    xpts, ypts = xpts[0:num], ypts[0:num]

    if isinstance(xpts[0], float):
        xpts = np.array(xpts)
        ypts = np.array(ypts)
    elif isinstance(xpts[0], type(1*u.cm)):
        xpts = np.array([i.value for i in xpts])*xpts[0].unit
        ypts = np.array([i.value for i in ypts])*ypts[0].unit
    else:
        raise TypeError

    return xpts, ypts



import numpy as np


class SurfaceMap:
    """Base class for surface maps."""
    def __init__(self, nlon=361, nlat=181, map=None):
        """
        
        :param nlon: Number of longitude bins
        :param nlat: Number of latitude bins
        :param map: If present, the surface map to use
        """



"""``interpu()``: 1D linear interpolation using astropy quantities.

This is a wrapper for numpy.interp for use when using astropy quantities. If
x and xp have different units, xp is converted to the units of x before
interpolation. An exception is raised if the units are not compatible
(i.e., the units of xp cannot be converted to the units of x).

:Author: Matthew Burger
"""
import numpy as np
import astropy.units as u


def interpu(x, xp, fp, **kwargs):
    """Return one dimensional interpolated astropy quantities.
    :param x: The x-coordinates at which to evaluate the interpolated values
    :param xp: The x-coordinates of the data points.
    :param fp: The y-coordinates of the data points
    :param kwargs:
    :return:
    
    **Notes**
    x and xp must have compatible units. See `numpy.interp
    <https://docs.scipy.org/doc/numpy/reference/generated/numpy.interp.html>`_
    for details on interpolation.
    """
    quantity = type(1 * u.km)
    
    if isinstance(x, quantity):
        x_ = x.value
    else:
        raise TypeError('x must be an astropy quantity')
    
    if isinstance(xp, quantity):
        if x.unit == xp.unit:
            xp_ = xp.value
        else:
            xp_ = xp.to(x.unit).value
    else:
        raise TypeError('xp must be an astropy quantity')

    if isinstance(fp, quantity):
        fp_ = fp.value
    else:
        raise TypeError('fp must be an astropy quantity')
    
    result = np.interp(x_, xp_, fp_, **kwargs)
    return result * fp.unit



import os
import pickle
import glob
from nexoclom import Input, __file__ as basefile

basepath = os.path.dirname(basefile)

inputfiles = glob.glob(os.path.join(os.path.dirname(__file__), 'test_data',
                                    'inputfiles', '*.input'))
inputs = []
for inputfile in inputfiles:
    input_ = Input(inputfile)

    # Correct the surfaceinteraction.mapfile path
    if ((input_.surfaceinteraction.sticktype == 'surface map') and
        (input_.surfaceinteraction.mapfile != 'default')):
        # put correct surfaceinteraction.mapfile path
        input_.surfaceinteraction.mapfile = os.path.join(basepath, 'modelcode',
            'tests', 'test_data', 'sticking_maps',
            os.path.basename(input_.surfaceinteraction.mapfile))
    else:
        pass
    
    # Correct the spatialdist.mapfile path
    if ((input_.spatialdist.type == 'surface map') and
        (input_.spatialdist.mapfile != 'default')):
        # put correct spatialdist.mapfile path
        input_.spatialdist.mapfile = os.path.join(basepath, 'modelcode', 'tests',
                                                  'test_data', 'surface_maps',
                                                  os.path.basename(input_.spatialdist.mapfile))
    else:
        pass
    
    inputs.append(input_)

with open('test_data/input_classes_data.pkl', 'wb') as f:
    pickle.dump((inputfiles, inputs), f)


import os.path
import numpy as np
import pandas as pd
import dask as dd
import pickle
import random
import copy
import astropy.units as u
from astropy.time import Time
from sklearn.neighbors import KDTree, BallTree
from scipy.interpolate import RectBivariateSpline
import scipy.linalg as lin
from psycopg2.extras import execute_batch
from nexoclom.math import fit_model

from nexoclom.modelcode import ModelResult
from nexoclom.utilities import database_connect
from nexoclom import Input
from nexoclom import Output


xcols = ['x', 'y', 'z']
borecols = ['xbore', 'ybore', 'zbore']
NLONBINS, NLATBINS, NVELBINS = 72, 36, 100


# Helper functions
def _should_add_weight(index, saved):
    return index in saved


def _add_weight(x, ratio):
    return np.append(x, ratio)


def _add_index(x, i):
    return np.append(x, i)


class InputError(Exception):
    """Raised when a required parameter is not included."""
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


class LOSResult(ModelResult):
    """Class to contain the LOS result from multiple outputfiles."""
    def __init__(self, scdata, quantity='radiance', dphi=1*u.deg):
        """Determine column or emission along lines of sight.
        This assumes the model has already been run.
        
        Parameters
        ==========
        scdata
            Spacecraft data object (currently designed for MESSENGERdata object
            but can be faked for other types of data)
            
        quantity
            Quantity to calculate: 'column', 'radiance', 'density'
            
        dphi
            Angular size of the view cone. Default = r deg.
        """
        format_ = {'quantity':quantity}
        super().__init__(format_, species=scdata.species)
        
        # Basic information
        self.scdata = scdata
        self.scdata.set_frame('Model')
        self.type = 'LineOfSight'
        self.unit = None
        self.dphi = dphi.to(u.rad).value
        self.oedge = None
        
        self.fitted = None
        self.outputfiles = []
        nspec = len(self.scdata)
        self.radiance = pd.Series(np.zeros(nspec), index=scdata.data.index)
        self.npackets = pd.Series(np.zeros(nspec), index=scdata.data.index,
                                  dtype=int)
        self.totalsource = 0.
        self.sourcemap = None
        self.modelfiles = {}

    def delete_models(self):
        """Deletes any LOSResult models associated with this data and input
        This may never actually do anything. Overwrite=True will also
        erase the outputfiles (which erases any models that depend on them).
        Unless I put separate outputfile and modelfile delete switches,
        This shouldn't do anything"""
        
        search_results = self.search()
        if len(search_results) != 0:
            print('Warning: LOSResult.delete_models found something to delete')
            for _, search_result in search_results.items():
                if search_result is not None:
                    idnum, modelfile = search_result
                    with database_connect() as con:
                        cur = con.cursor()
                        cur.execute(f'''DELETE from uvvsmodels
                                       WHERE idnum = %s''', (idnum, ))
                    if os.path.exists(modelfile):
                        os.remove(modelfile)
        else:
            pass

    def save(self, data_packets):
        # Insert the model into the database
        # Save is on an outputfile basis
        if self.quantity == 'radiance':
            mech = ', '.join(sorted([m for m in self.mechanism]))
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ', '.join([str(w) for w in wave_])
        else:
            mech = None
            wave = None

        # Save query with all white space removed and lowercase
        tempname = f'temp_{str(random.randint(0, 1000000))}'

        with database_connect() as con:
            cur = con.cursor()
            cur.execute(f'''INSERT into uvvsmodels (out_idnum, quantity,
                            query, dphi, mechanism, wavelength,
                            fitted, filename)
                            values (%s, %s, %s, %s, %s, %s, %s, %s)''',
                        (data_packets.out_idnum, self.quantity,
                         self.scdata.query, self.dphi, mech, wave, self.fitted,
                         tempname))

            # Determine the savefile name
            idnum_ = pd.read_sql(f'''SELECT idnum
                                     FROM uvvsmodels
                                     WHERE filename='{tempname}';''', con)
            assert len(idnum_) == 1
            idnum = int(idnum_.idnum[0])

            savefile = os.path.join(os.path.dirname(data_packets.outputfile),
                                    f'model.{idnum}.pkl')

            cur.execute(f'''UPDATE uvvsmodels
                            SET filename=%s
                            WHERE idnum=%s''', (savefile, idnum))
            
        with open(savefile, 'wb') as f:
            pickle.dump(data_packets, f)
            
        return savefile

    def search(self):
        """
        :return: dictionary containing search results:
                 {outputfilename: (modelfile_id, modelfile_name)}
        """
        search_results = {}
        for outputfile in self.outputfiles:
            with database_connect() as con:
                # Determine the id of the outputfile
                idnum_ = pd.read_sql(
                    f'''SELECT idnum
                        FROM outputfile
                        WHERE filename='{outputfile}' ''', con)
                oid = idnum_.idnum[0]
            
                if self.quantity == 'radiance':
                    mech = ("mechanism = '" +
                            ", ".join(sorted([m for m in self.mechanism])) +
                            "'")
                    wave_ = sorted([w.value for w in self.wavelength])
                    wave = ("wavelength = '" +
                            ", ".join([str(w) for w in wave_]) +
                            "'")
                else:
                    mech = 'mechanism is NULL'
                    wave = 'wavelength is NULL'
            
                result = pd.read_sql(
                    f'''SELECT idnum, filename FROM uvvsmodels
                        WHERE out_idnum={oid} and
                              quantity = '{self.quantity}' and
                              query = '{self.scdata.query}' and
                              dphi = {self.dphi} and
                              {mech} and
                              {wave} and
                              fitted = {self.fitted}''', con)
            
                # Should only have one match per outputfile
                assert len(result) <= 1
                
                if len(result) == 0:
                    search_results[outputfile] = None
                else:
                    search_results[outputfile] = (result.iloc[0, 0],
                                                  result.iloc[0, 1])
                
        return search_results
    
    @staticmethod
    def restore(modelfile):
        # Restore is on an outputfile basis
        with open(modelfile, 'rb') as f:
            data_packets = pickle.load(f)

        return data_packets
    
    def _data_setup(self):
        # distance of s/c from planet
        data = self.scdata.data
        dist_from_plan = np.sqrt(data.x**2 + data.y**2 + data.z**2)
    
        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore - data.y*data.ybore -
                         data.z*data.zbore)/dist_from_plan)
        
        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan.loc[ang > asize_plan] = 1e30
        
        return dist_from_plan
    
    def _spectrum_process(self, spectrum, packets, tree, dist, i, ofile):
        x_sc = spectrum[xcols].values.astype(float)
        bore = spectrum[borecols].values.astype(float)
    
        dd = 30  # Furthest distance we need to look
        x_far = x_sc + bore * dd
        while np.linalg.norm(x_far) > self.oedge:
            dd -= 0.1
            x_far = x_sc + bore * dd
    
        t = [0.05]
        while t[-1] < dd:
            t.append(t[-1] + t[-1] * np.sin(self.dphi))
        t = np.array(t)
        Xbore = x_sc[np.newaxis, :] + bore[np.newaxis, :] * t[:, np.newaxis]
    
        wid = t * np.sin(self.dphi)
        ind = np.concatenate(tree.query_radius(Xbore, wid))
        ilocs = np.unique(ind).astype(int)
        indicies = packets.iloc[ilocs].index
        subset = packets.loc[indicies]
    
        xpr = subset[xcols] - x_sc[np.newaxis, :]
        rpr = np.sqrt(xpr['x'] * xpr['x'] +
                      xpr['y'] * xpr['y'] +
                      xpr['z'] * xpr['z'])
    
        losrad = np.sum(xpr * bore[np.newaxis, :], axis=1)
        inview = rpr < dist
        
        if np.any(inview):
            used_packets = subset[inview].index.to_list()
            used_packets0 = subset.loc[inview, 'Index']

            Apix = np.pi * (rpr[inview] * np.sin(self.dphi))**2 * (
                self.unit.to(u.cm))**2
            wtemp = subset.loc[inview, 'weight'] / Apix
            if self.quantity == 'radiance':
                # Determine if any packets are in shadow
                # Projection of packet onto LOS
                # Point along LOS the packet represents
                losrad_ = losrad[inview].values
                hit = (x_sc[np.newaxis, :] +
                       bore[np.newaxis, :] * losrad_[:, np.newaxis])
                rhohit = np.linalg.norm(hit[:, [0, 2]], axis=1)
                out_of_shadow = (rhohit > 1) | (hit[:, 1] < 0)
                wtemp *= out_of_shadow

                rad = wtemp.sum()
                # Save the weight information
                rat = spectrum.radiance/rad if rad > 0 else 0.
                
                params = []
                for U, w, f0, u0 in zip(used_packets, wtemp.values,
                                        packets.loc[used_packets, 'frac0'].values,
                                        used_packets0.values):
                    params.append((self.scdata.query, ofile, int(i), int(U), w,
                                   float(f0), int(u0), rat))
                # params = list(zip(used_packets, wtemp.values,
                #                   packets.loc[used_packets, 'frac0'].values.astype('float'),
                #                   used_packets0.values.astype('float')))

                statement = '''INSERT into savedpackets (
                                   query, outputfile, specind, oint,
                                   weight, frac0, index0, ratio) VALUES (
                                   %s, %s, %s, %s, %s::DOUBLE PRECISION,
                                   %s::DOUBLE PRECISION, %s,
                                   %s::DOUBLE PRECISION)'''
                with database_connect() as con:
                    cur = con.cursor()
                    execute_batch(cur, statement, params)
                del params
            else:
                assert False, 'Other quantities not set up.'
        else:
            pass
        
    @staticmethod
    def _tree(values, treetype='KDTree'):
        if treetype == 'KDTree':
            return KDTree(values)
        elif treetype == 'BallTree':
            return BallTree(values)
        
    def determine_source_from_data(self, modnum, weight_method='scaling'):
        modkey = f'model{modnum:02d}'
        maskkey = f'mask{modnum:02d}'
        mask = self.scdata.data[maskkey]
        
        self.outputfiles = self.scdata.model_info[modkey]['outputfiles']
        surface_source = pd.DataFrame(columns=['outputfile', 'Index', 'longitude',
                                               'latitude', 'velocity'])
        nsteps, endtime = 0, 0
        index0 = {}
        for outputfile in self.outputfiles:
            # Restore the unfit output file
            output = Output.restore(outputfile)
            self.totalsource += output.totalsource
            endtime = output.inputs.options.endtime
            nsteps = output.nsteps
            
            with database_connect() as con:
                ind0 = pd.read_sql(
                    f'''SELECT DISTINCT index0 FROM savedpackets
                        WHERE query='{self.scdata.query}' and
                              outputfile='{outputfile}';''', con)
            index0[outputfile] = ind0.index0.to_list()
            del output

            # vel_ = np.sqrt(output.X0.vx**2 + output.X0.vy**2 +
            #                output.X0.vz**2) * output.inputs.geometry.planet.radius
            # surface_source = surface_source.append(pd.DataFrame(
            #     {'outputfile': [outputfile for _ in range(len(ind0))],
            #      'Index': ind0,
            #      'longitude': output.X0.loc[ind0, 'longitude'].values,
            #      'latitude': output.X0.loc[ind0, 'latitude'].values,
            #      'velocity': vel_.loc[ind0]}), ignore_index = True)
            
        if weight_method == 'scaling':
            # Determine the proper weightings
            outputfile_str = self.outputfiles.__str__().replace('[', '(').replace(']', ')')
            statement = f'''UPDATE savedpackets
                            SET scale_factor=%s::DOUBLE PRECISION
                            WHERE query='{self.scdata.query}' and
                                outputfile=%s and
                                index0=%s'''
            with database_connect() as con:
                cur = con.cursor()
                scale_factor = pd.read_sql(
                    f'''SELECT outputfile, index0, AVG(ratio) from savedpackets
                        WHERE query='{self.scdata.query}' and
                              outputfile in {outputfile_str}
                        GROUP BY outputfile, index0''', con)
                
                scale_factor_ind = scale_factor.set_index(['outputfile', 'index0'])
                params = list(zip(scale_factor_ind['avg'].values,
                              [ind[0] for ind in scale_factor_ind.index],
                              [ind[1] for ind in scale_factor_ind.index]))
                execute_batch(cur, statement, params)
                # for ind, scfactor in scale_factor_ind.iterrows():
                    # cur.execute(f'''UPDATE savedpackets
                    #                 SET scale_factor=%s::DOUBLE PRECISION
                    #                 WHERE query='{self.scdata.query}' and
                    #                       outputfile=%s and
                    #                       index0=%s''',
                    #             (scfactor, ind[0], ind[1]))

                weight = pd.read_sql(
                    f'''SELECT outputfile, oint, index0, weight*scale_factor fit_weight
                        FROM savedpackets
                        WHERE query='{self.scdata.query}' and
                              outputfile in {outputfile_str};''', con)
            
            from IPython import embed; embed()
            import sys; sys.exit()
            
            # ind = pd.MultiIndex.from_arrays([weight.outputfile.to_list(),
            #                                  weight.index0.to_list()])
            # scale_factor_ = scale_factor_ind.loc[ind, 'avg'].reset_index()
            # fit_weight = weight.weight * scale_factor_['avg']
            
            # Determine surface weighting
            # surface_ind = surface_source.set_index(['outputfile', 'Index'])
            # surface_ind.loc[scale_factor_.index, 'weight'] = scale_factor_
            # surface_ind.dropna(inplace=True)
            # surface_source = surface_ind.reset_index()
        elif weight_method == 'leastsq':
            # This method produces too many scale_factors < 0
            assert 0
            
            # Solve linear equations to get best fit
            temp = self.data_packets.data.set_index('specind')
            temp = temp.loc[mask]
            temp.reset_index(inplace=True)
            
            W = np.zeros((len(data), len(self.modelfiles),
                          self.data_packets.data.Index0.max()+1))
            
            spec_ind = [data.index.get_loc(i) for i in temp.specind.values]
            pack_ind = temp.index.to_list()
            
            outputfiles = {ofile: i for i, ofile
                           in enumerate(self.modelfiles.keys())}
            oo = [outputfiles[ofile] for ofile in temp.outputfile]
            W[spec_ind, oo, temp.Index0.to_list()] = temp.weight
            
            W2 = W.squeeze(axis=1)/W.mean()
            # assert np.all(W == W2.reshape(W.shape))
            assert W.shape[1] == 1
            scale_factor_, r, rank, s = lin.lstsq(W2, data.loc[:, 'radiance'].values)
            scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
            self.data_packets.data['scale_factor'] = (
                scale_factor_[self.data_packets.data.Index0.to_list()])
            self.data_packets.data['fit_weight'] = (self.data_packets.data.weight *
                                                    self.data_packets.data.scale_factor)

            surface_source['weight'] = scale_factor_[surface_source.Index.to_list()]
            
            scaled = pd.read_pickle('surface_source_scaled.pkl')

            from IPython import embed; embed()
            import sys; sys.exit()
            
    
            self.data_packets.data['scale_factor'] = (
                scale_factor_[self.data_packets.data['Index0'].to_list()])

            # Determine surface weighting
            
            
            
            surface_source.dropna(inplace=True)
            # surface_source = surface_source[surface_source.weight > 0]
        else:
            raise InputError('LOSResult.determine_source_from_data: '
                             'Not a valid weighting method')

        initial = set(zip(self.data_packets.data.outputfile.to_list(),
                          self.data_packets.data.Index0.to_list(),
                          self.data_packets.data.frac0.to_list(),
                          self.data_packets.data.scale_factor.to_list()))
        frac0 = sum(x[2] * x[3] for x in initial)
        self.totalsource = frac0 * nsteps
        mod_rate = self.totalsource / endtime.value
        atoms_per_packet = 1e23 / mod_rate

        spec_group = self.data_packets.data.groupby('specind')
        fitrad_ = spec_group['fit_weight'].sum() * atoms_per_packet

        self.radiance.loc[fitrad_.index] = fitrad_
        self.radiance *= u.R

        # Create a new sourcemap
        source, xx, yy = np.histogram2d(surface_source['longitude'],
                                        surface_source['latitude'],
                                        weights=surface_source['weight'],
                                        range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                        bins=(NLONBINS, NLATBINS))
        source = source/np.cos(yy+(yy[1]-yy[0])/2.)[np.newaxis,:-1]
        source[:,[0,-1]] = 0
        
        v_source, v = np.histogram(surface_source['velocity'], bins=NVELBINS,
                                   range=[0, surface_source['velocity'].max()],
                                   weights=surface_source['weight'])
        v_source /= np.max(v_source)

        # packets available
        packets, _, _ = np.histogram2d(surface_source['longitude'],
                                       surface_source['latitude'],
                                       range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
                                       bins=(NLONBINS, NLATBINS))
        packets = packets/np.cos(yy+(yy[1]-yy[0])/2.)[np.newaxis,:-1]
        packets[:,[0,-1]] = 0
        v_packets, _ = np.histogram(surface_source['velocity'], bins=NVELBINS,
                                    range=[0, surface_source['velocity'].max()])
        v_packets = v_packets / np.max(v_packets)
        
        sourcemap = {'longitude': xx*u.rad,
                     'latitude': yy*u.rad,
                     'abundance': source,
                     'p_available': packets,
                     'velocity': v*u.km/u.s,
                     'vdist': v_source,
                     'v_available': v_packets,
                     'coordinate_system': 'solar-fixed'}
        self.sourcemap = sourcemap
        
    def simulate_data_from_inputs(self, inputs_, npackets, overwrite=False,
                                  packs_per_it=None):
        """Given a set of inputs, determine what the spacecraft should see.
        
        Parameters
        ==========
        inputs_
            A nexoclom Input object or the name of an inputs file
        """
        if isinstance(inputs_, str):
            self.inputs = Input(inputs_)
        elif isinstance(inputs_, Input):
            self.inputs = copy.deepcopy(inputs_)
        else:
            raise InputError('nexoclom.LOSResult', 'Problem with the inputs.')
    
        # TAA needs to match the data
        self.inputs.geometry.taa = self.scdata.taa
        self.unit = u.def_unit('R_' + self.inputs.geometry.planet.object,
                               self.inputs.geometry.planet.radius)
    
        # If using a planet-fixed source map, need to set subsolarlon
        if ((self.inputs.spatialdist.type == 'surface map') and
            (self.inputs.spatialdist.coordinate_system == 'planet-fixed')):
            self.inputs.spatialdist.subsolarlon = self.scdata.subslong.median() * u.rad
        else:
            pass
    
        # Run the model
        self.fitted = False
        self.inputs.run(npackets, packs_per_it=packs_per_it, overwrite=overwrite)
        self.search_for_outputs()
        
        dist_from_plan = self._data_setup()
        data = self.scdata.data
        
        for outputfile in self.outputfiles:
            with database_connect() as con:
                ct = pd.read_sql(
                    f'''SELECT count(*) FROM savedpackets
                        WHERE query='{self.scdata.query}' and
                              outputfile='{outputfile}';''', con)
            output = Output.restore(outputfile)
            self.totalsource += output.totalsource
            if ct.loc[0, 'count'] == 0:
                packets = copy.deepcopy(output.X)
                packets['frac0'] = output.X0.loc[packets['Index'], 'frac'].values
                packets['radvel_sun'] = (packets['vy'] +
                                         output.vrplanet.to(self.unit / u.s).value)
                self.oedge = output.inputs.options.outeredge * 2
        
                # Will base shadow on line of sight, not the packets
                out_of_shadow = np.ones(len(packets))
                self.packet_weighting(packets, out_of_shadow, output.aplanet)
        
                # This sets limits on regions where packets might be
                tree = self._tree(packets[xcols].values)
        
                print(f'{data.shape[0]} spectra taken.')
                for i, spectrum in data.iterrows():
                    self._spectrum_process(spectrum, packets, tree,
                                           dist_from_plan[i],
                                           i, outputfile)
                    
                    ind = data.index.get_loc(i)
                    if (ind % (len(data)//10)) == 0:
                        print(f'Completed {ind+1} spectra')
                    else:
                        pass

            del output

        mod_rate = self.totalsource / self.inputs.options.endtime.value
        atoms_per_packet = 1e23 / mod_rate
        
        outputfile_str = self.outputfiles.__str__().replace('[', '(').replace(']',')')
        with database_connect() as con:
            rad_ = pd.read_sql(
                f'''SELECT specind, sum(weight) from savedpackets
                    WHERE query='{self.scdata.query}' and
                          outputfile in {outputfile_str}
                    GROUP BY specind''', con)
        self.radiance.loc[rad_.specind.to_list()] = rad_['sum'].values * atoms_per_packet
        self.radiance *= u.R

    def recompute_radiance(self, modnum, new_source):
        data = self.scdata.data
        radiance = pd.Series(index=data.index)

        modkey = f'model{modnum:02d}'
        maskkey = f'mask{modnum:02d}'
        
        self.modelfiles = self.scdata.model_info[modkey]['modelfiles']
        assert len(self.modelfiles) > 0
        self.outputfiles = self.modelfiles.keys()

        self.data_packets = DataPackets()
        surface_source = pd.DataFrame(columns=['outputfile', 'Index', 'longitude',
                                               'latitude'])
        nsteps, endtime = 0, 0
        for outputfile, modefile in self.modelfiles.items():
            # Restore the unfit output file
            output = Output.restore(outputfile)
            self.totalsource += output.totalsource
            endtime = output.inputs.options.endtime
            nsteps = output.nsteps
            data_packets_it = self.restore(modefile)
            
            self.data_packets.data = self.data_packets.data.append(
                data_packets_it.data)
            self.data_packets.totalsource += data_packets_it.totalsource
            
            ind0 = data_packets_it.data.Index0.unique()
            surface_source = surface_source.append(pd.DataFrame(
                {'outputfile': [outputfile for _ in range(len(ind0))],
                 'Index': ind0,
                 'longitude': output.X0.loc[ind0, 'longitude'].values,
                 'latitude': output.X0.loc[ind0, 'latitude'].values}), ignore_index=True)

            del output
            del data_packets_it

        # Determine new weighting from given map
        new_weight_fn = RectBivariateSpline(new_source['longitude'][:-1].value,
                                            np.sin(new_source['latitude'][:-1].value),
                                            new_source['abundance'])
        scale_factor_ = new_weight_fn(surface_source.longitude,
                                      np.sin(surface_source.latitude),
                                      grid=False)
        scale_factor_ /= scale_factor_[scale_factor_ > 0].mean()
        surface_source['scale_factor'] = scale_factor_
        
        surface_ind = surface_source.set_index(['outputfile', 'Index'])
        ind0_ind = self.data_packets.data.set_index(['outputfile', 'Index0'])
        ind0_ind.loc[surface_ind.index, 'scale_factor'] = surface_ind.scale_factor
        # ind0_ind = ind0_ind.dropna()
        
        self.data_packets.data = ind0_ind.reset_index()
        self.data_packets.data['new_weight'] = (self.data_packets.data.weight *
                                                self.data_packets.data.scale_factor)
        
        initial = set(zip(self.data_packets.data.outputfile.to_list(),
                          self.data_packets.data.Index0.to_list(),
                          self.data_packets.data.frac0.to_list(),
                          self.data_packets.data.scale_factor.to_list()))
        frac0 = sum(x[2] * x[3] for x in initial)
        self.totalsource = frac0 * nsteps
        mod_rate = self.totalsource / endtime.value
        atoms_per_packet = 1e23 / mod_rate

        spec_group = self.data_packets.data.groupby('specind')
        fitrad_ = spec_group['new_weight'].sum() * atoms_per_packet
        # fitrad = spec_group['weight'].sum() * atoms_per_packet

        self.radiance.loc[fitrad_.index] = fitrad_
        self.radiance *= u.R


# if __name__ == '__main__':
#     import matplotlib.pyplot as plt
#     hist, xx, yy = np.histogram2d(surface_ind.longitude, surface_ind.latitude,
#                                   weights=surface_ind.scale_factor,
#                                   range=[[0, 2*np.pi], [-np.pi/2, np.pi/2]],
#                                   bins=(NLONBINS, NLATBINS))
#     hist = hist / np.cos(yy + (yy[1] - yy[0]) / 2.)[np.newaxis, :-1]
#     hist *= new_source['abundance'].mean()/hist.mean()
#
#     fig, ax = plt.subplots(1, 2)
#     ax[0].imshow(new_source['abundance'], vmin=0, vmax=100)
#     ax[1].imshow(hist, vmin=0, vmax=100)
#     plt.show()
#     #
#
#     plt.plot(np.sum(new_source['abundance'], axis=0))
#     plt.plot(np.sum(hist, axis=0))
#     plt.show()
#     plt.plot(np.sum(new_source['abundance'], axis=1))
#     plt.plot(np.sum(hist, axis=1))
#     plt.show()
#     #
#     # from IPython import embed; embed()
#     # import sys; sys.exit()



"""``rotation_matrix()``: Compute the rotation matrix about an axis."""
import numpy as np


def rotation_matrix(theta, axis):
    """Compute the rotation matrix for a rotation of theta about axis."""
    unit_vec = axis/np.linalg.norm(axis)
    lx, ly, lz = unit_vec[0], unit_vec[1], unit_vec[2]
    c, s = np.cos(theta), np.sin(theta)
    rot_mat = np.asmatrix([[lx**2+(1-lx**2)*c, lx*ly*(1-c)+lz*s, lx*lz*(1-c)-ly*s],
                          [lx*ly*(1-c)-lz*s, ly**2+(1-ly**2)*c, ly*lz*(1-c)+lx*s],
                          [lx*lz*(1-c)+ly*s, ly*lz*(1-c)-lx*s, lz**2+(1-lz**2)*c]])

    return rot_mat



from nexoclom.modelcode.Input import Input
from nexoclom.modelcode.Output import Output
from nexoclom.modelcode.LOSResult import LOSResult
from nexoclom.modelcode.ModelImage import ModelImage
from nexoclom.utilities.configure import configure
from nexoclom.solarsystem import SSObject


__name__ = 'nexoclom'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'
__version__ = '3.2.2'
__date__ = '2022-05-04'

configure(verbose=True)



"""Create an object for each Solar System body containing basic information.
Information stored:
* object: Object name
* orbits: Body that the object orbits
* radius: in km
* mass: in kg
* a: orbital semi major axis. In AU if orbits the Sun; km
    if orbits a planet
* e: orbital eccentricity
* tilt: tilt of planetary axis in degrees
* rotperiod: rotation period in hours
* orbperiod: orbital period in days
* GM: mass * G in m**3/s**2
* moons: returned as a list of SSObjects

Values are astropy units quantities when appropriate.

"""
import os
import pandas as pd
from astropy import constants as const
from astropy import units as u
from nexoclom import __file__ as basefile


class SSObject:
    """Creates Solar System object."""
    def __init__(self, obj):
        basepath = os.path.dirname(basefile)
        pklfile = os.path.join(basepath, 'data', 'PlanetaryConstants.pkl')
        if not os.path.exists(pklfile):
            set_up_planetary_constants(pklfile)
        else:
            pass

        constants = pd.read_pickle(pklfile)
        
        row = constants.loc[constants.Object.apply(lambda x: x.casefold()) ==
                            obj.casefold()]
        if len(row) == 1:
            row = row.iloc[0]

            self.object = row.Object
            self.orbits = row.orbits
            self.radius = row.radius * u.km
            self.mass = row.mass * u.kg
            self.a = row.a
            self.e = row.e
            self.tilt = row.tilt * u.deg
            self.rotperiod = row.rot_period * u.h
            self.orbperiod = row.orb_period * u.d
            self.GM = -self.mass * const.G

            self.moons = [SSObject(moon) for moon in
                constants.loc[constants.orbits == self.object, 'Object'].to_list()]
            if len(self.moons) == 0:
                self.moons = None
            else:
                pass

            if self.orbits == 'Milky Way':
                self.type = 'Star'
                self.a *= u.km
            elif self.orbits == 'Sun':
                self.type = 'Planet'
                self.a *= u.au
            else:
                self.type = 'Moon'
                self.a *= u.km
        else:
            print(f'Object {obj} does not exist in table.')
            self.object = None

    def __len__(self):
        # Returns number of objects (e.g. Planet + moons) in the SSObeject
        return 1 if self.moons is None else len(self.moons)+1

    def __eq__(self, other):
        return self.object == other.object

    def __hash__(self):
        return hash((self.object, ))
    
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        out = (f'Object: {self.object}\n'
               f'Type = {self.type}\n'
               f'Orbits {self.orbits}\n'
               f'Radius = {self.radius:0.2f}\n'
               f'Mass = {self.mass:0.2e}\n'
               f'a = {self.a:0.2f}\n'
               f'Eccentricity = {self.e:0.2f}\n'
               f'Tilt = {self.tilt:0.2f}\n'
               f'Rotation Period = {self.rotperiod:0.2f}\n'
               f'Orbital Period = {self.orbperiod:0.2f}\n'
               f'GM = {self.GM:0.2e}')
        return out
    
def set_up_planetary_constants(pklfile):
    basepath = os.path.dirname(basefile)
    objfile = os.path.join(basepath, 'data', 'SolarSystemContents.csv')
    objects = pd.read_csv(objfile)
    
    # Columns to add: radius, mass, a, e, tilt, rot_period, orb_period
    
    
    table = pd.read_csv(pklfile.replace('.pkl', '.dat'), skipinitialspace=True,
                        skip_blank_lines=True, comment='#', sep=':')
    table.columns = [col.strip() for col in table.columns]
    table.Object = table.Object.apply(lambda x: x.strip())
    table.orbits = table.orbits.apply(lambda x: x.strip())
    table.to_pickle(pklfile)



'''Exceptions used in nexoclom'''
class InputError(Exception):
    """Raised when a required parameter is not included"""
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message

class ConfigfileError(Exception):
    def __init__(self, configfile, missing):
        self.expression = configfile
        self.message = f'{missing} not defined in {configfile}'


